;; Generated package description from breadcrumb.el  -*- no-byte-compile: t -*-
(define-package "breadcrumb" "1.0.1.0.20250208.145328" "project and imenu-based breadcrumb paths" '((emacs "28.1") (project "0.9.8")) :commit "da34d030e6d01db2bba45b30080204b23a714c9f" :url "https://elpa.gnu.org/packages/breadcrumb.html" :authors '(("João Távora" . "joaotavora@gmail.com")) :maintainer '("João Távora" . "joaotavora@gmail.com"))
